/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description
    Splits mesh by making internal faces external. Uses attachDetach.

    Generates a meshModifier of the form:

    Splitter
    {
        type                       attachDetach;
        faceZoneName               membraneFaces;
        masterPatchName            masterPatch;
        slavePatchName             slavePatch;
        triggerTimes               runTime.value();
    }

    so will detach at the current time and split all faces in membraneFaces
    into masterPatch and slavePatch (which have to be present but of 0 size)

    Preserves additional faceSets. The faceSet that was used to split the mesh is erased.
    Assumes that the cells are not renumbered

\*---------------------------------------------------------------------------*/

#include "argList.H"
#include "Time.H"
#include "polyTopoChange.H"
#include "morphMesh.H"
#include "mapPolyMesh.H"
#include "faceSet.H"
#include "Map.H"
#include "ListOps.H"
#include "meshTools.H"
#include "attachDetach.H"
#include "attachPolyMesh.H"
#include "IndirectList.H"
#include "regionSide.H"

using namespace Foam;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

// Calculation engine for set of faces in a mesh
typedef PrimitivePatch<face, List, const pointField&> facePatch;


// Find edge between points v0 and v1.
label findEdge(const primitiveMesh& mesh, const label v0, const label v1)
{
    const labelList& pEdges = mesh.pointEdges()[v0];

    forAll(pEdges, pEdgeI)
    {
        label edgeI = pEdges[pEdgeI];

        const edge& e = mesh.edges()[edgeI];

        if (e.otherVertex(v0) == v1)
        {
            return edgeI;
        }
    }

    FatalErrorIn
    (
        "findEdge(const primitiveMesh&, const label, const label)"
    )   << "Cannot find edge between mesh points " << v0 << " and " << v1
        << abort(FatalError);

    return -1;
}


// Checks whether patch present
void checkPatch(const polyBoundaryMesh& bMesh, const word& name)
{
    label patchI = bMesh.findPatchID(name);

    if (patchI == -1)
    {
        FatalErrorIn("checkPatch(const polyBoundaryMesh&, const word&)")
            << "Cannot find patch " << name << endl
            << "It should be present but of zero size" << endl
            << "Valid patches are " << bMesh.names()
            << exit(FatalError);
    }

    if (bMesh[patchI].size() != 0)
    {
        FatalErrorIn("checkPatch(const polyBoundaryMesh&, const word&)")
            << "Patch " << name << " is present but not of zero size"
            << exit(FatalError);
    }
}

wordList getFaceSets(const polyMesh &m) {
    DynamicList<word> sets;
    fileName setDir=m.rootPath()/m.caseName()/"constant"/polyMesh::meshSubDir/"sets";

    fileNameList files=readDir(setDir);

    forAll(files,i) {
        IOobject f
            (
                files[i],
                "constant",
                polyMesh::meshSubDir/"sets",
                m,
                IOobject::MUST_READ,
                IOobject::NO_WRITE
                
            );
        f.headerOk();
        if(f.headerClassName()=="faceSet") {
            sets.append(files[i]);
        }
    }

    sets.shrink();

    return sets;
}

// Main program:

int main(int argc, char *argv[])
{
    Foam::argList::noParallel();

    Foam::argList::validArgs.append("faceSet");
    Foam::argList::validArgs.append("masterPatch");
    Foam::argList::validArgs.append("slavePatch");

#   include "setRootCase.H"
#   include "createTime.H"

    word setName(args.args()[3]);
    word masterPatch(args.args()[4]);
    word slavePatch(args.args()[5]);

    attachPolyMesh pMesh
    (
        IOobject
        (
            attachPolyMesh::defaultRegion,
            runTime.timeName(),
            runTime
        )
    );

    wordList theFaceSets=getFaceSets(pMesh);
    //   Info << "Face sets: " << theFaceSets << endl;

    // List of faces to split
    faceSet facesSet(pMesh, setName);

    if(facesSet.size() > 0) {
        Info<< "Read " << facesSet.size() << " faces to split" << endl << endl;
    } else {
            FatalErrorIn(args.executable())
            << "FaceSet " << setName
            << " is empty."
            << exit(FatalError);
    }

    // Convert into labelList and check

    labelList faces(facesSet.toc());

    forAll(faces, i)
    {
        if (!pMesh.isInternalFace(faces[i]))
        {
            FatalErrorIn(args.executable())
            << "Face " << faces[i] << " in faceSet " << setName
            << " is not an internal face."
            << exit(FatalError);
        }
    }


    // Check for empty master and slave patches
    checkPatch(pMesh.boundaryMesh(), masterPatch);
    checkPatch(pMesh.boundaryMesh(), slavePatch);


    //
    // Find 'side' of all faces on splitregion. Uses regionSide which needs
    // set of edges on side of this region. Use PrimitivePatch to find these.
    //

    IndirectList<face> zoneFaces(pMesh.faces(), faces);

    // Addressing on faces only in mesh vertices.
    facePatch fPatch(zoneFaces(), pMesh.points());

    const labelList& meshPoints = fPatch.meshPoints();

    // Mark all fence edges : edges on boundary of fPatch but not on boundary
    // of polyMesh
    labelHashSet fenceEdges(fPatch.size());

    const labelListList& allEdgeFaces = fPatch.edgeFaces();

    forAll(allEdgeFaces, patchEdgeI)
    {
        if (allEdgeFaces[patchEdgeI].size() == 1)
        {
            const edge& e = fPatch.edges()[patchEdgeI];

            label edgeI =
                findEdge
                (
                    pMesh,
                    meshPoints[e.start()],
                    meshPoints[e.end()]
                );

            fenceEdges.insert(edgeI);
        }
    }

    // Find sides reachable from 0th face of faceSet
    label startFaceI = faces[0];

    regionSide regionInfo
    (
        pMesh,
        facesSet,
        fenceEdges,
        pMesh.faceOwner()[startFaceI],
        startFaceI
    );

    // Determine flip state for all faces in faceSet
    boolList zoneFlip(faces.size());

    forAll(faces, i)
    {
        zoneFlip[i] = !regionInfo.sideOwner().found(faces[i]);
    }


    // Create and add face zones and mesh modifiers
    List<pointZone*> pz(0);
    List<faceZone*> fz(1);
    List<cellZone*> cz(0);

    fz[0] =
        new faceZone
        (
            "membraneFaces",
            faces,
            zoneFlip,
            0,
            pMesh.faceZones()
        );

    HashTable<faceSet*,word> theSets;

    HashTable<labelPairList*,word> cellInfo;

    // Get information about the cells on each side of the faceSet
    {
        forAll(theFaceSets,i) {
            if(theFaceSets[i]!=setName) {
                word name=theFaceSets[i];
                Info << "Adding zone for faceSet " << name << endl;
                theSets.insert(name,new faceSet(pMesh,name));
                faceSet &thisSet=*theSets[name];
                labelList theFaces(thisSet.toc());
                cellInfo.insert(name,new labelPairList(theFaces.size()));
                labelPairList &faceCells=*cellInfo[name];
                const labelList& own=pMesh.faceOwner();
                const labelList& neigh=pMesh.faceNeighbour();
                forAll(theFaces,faceI) {
                    label ind=theFaces[faceI];
                    faceCells[faceI].first()=own[ind];
                    faceCells[faceI].second()=neigh[ind];
                }
            }
        }
    }

    Info << "Adding point and face zones" << endl;
    pMesh.addZones(pz, fz, cz);


    List<polyMeshModifier*> tm(1);

    // Add the sliding interface mesh modifier to start working at current
    // time
    tm[0] =
        new attachDetach
        (
            "Splitter",
            0,
            pMesh,
            "membraneFaces",
            masterPatch,
            slavePatch,
            scalarField(1, runTime.value())
        );

    Info<< nl << "Constructed topologyModifier:" << endl;
    tm[0]->writeDict(Info);

    Info<< nl << "Adding topology modifiers" << endl;
    pMesh.addTopologyModifiers(tm);

    runTime++;

    pMesh.attach(false);

    // Reconstruct the faceSets from the cellNumbers 
    {
         const labelList& own=pMesh.faceOwner();
         const labelList& neigh=pMesh.faceNeighbour();
         const cellList &cells=pMesh.cells();

         wordList names=theSets.toc();
         forAll(names,i) {
             word name=names[i];

             labelPairList &faceCells=*cellInfo[name];
             faceSet &thisSet=*theSets[name];
             labelList theFaces(thisSet.toc());

             Info << "Resetting faceSet " << name << endl;
             thisSet.clear();

             forAll(faceCells,j) {
                 label cellA=faceCells[j].first();
                 label cellB=faceCells[j].second();
                 const cell &c=cells[cellA];
                 label foundFace=-1;

                 forAll(c,k) {
                     label fI=c[k];
                     if( (own[fI]==cellA && neigh[fI]==cellB) || (own[fI]==cellB && neigh[fI]==cellA) ) {
                         foundFace=fI;
                         break;
                     }
                 }
                 if(foundFace<0) {
                     //                     FatalErrorIn(args.executable())
                     Info    << " No matching face between cells " << faceCells[j] << " in set " << name << "found!";
                         //                             << exit(FatalError);
                 } else {
                     thisSet.insert(foundFace);
                 }
             }
         }
     }

    pMesh.removeFaceZones();

    Info << nl << "Writing polyMesh" << endl;
    if (!pMesh.write())
    {
        FatalErrorIn(args.executable())
            << "Failed writing polyMesh."
            << exit(FatalError);
    }

    Info << "Erasing face set " << setName << endl;
    facesSet.clear();
    facesSet.write();

    // Write the reconstructed faceSets
    {
        wordList names=theSets.toc();

        forAll(names,i) {
            word name=names[i];
            Info << "Writing faceSet " << name << " ( " << theSets[name]->size() << " ) " << endl;
            theSets[name]->write();
        }
    }

    Info << nl << "end" << endl;
    return 0;
}


// ************************************************************************* //
